from .basefolder import BaseAnVILFolder
from .workspace import Workspace

class Namespace(BaseAnVILFolder):
    def __init__(self, namespace_name):
        super().__init__(namespace_name)

    def fetch_workspace(self, workspace_name):
        ws = Workspace(self, workspace_name)
        self[ws.name] = ws
        return ws

    def __str__(self):
        out = "<Namespace {}>".format(self.name)
        for ws in self.workspaces:
            out += "\n{}".format(str(ws))


try:
    from anvilfs.anvilfs import AnVILFS
except ImportError as e:
    AnVILFS = None
from ._pyfilesystem2 import PyFilesystem2FilesSource
class AnVILFilesSource(PyFilesystem2FilesSource):
    plugin_type = 'anvil'
    required_module = AnVILFS
    required_package = "fs.anvilfs"
    def _open_fs(self, user_context):
        props = self._serialization_props(user_context)
        handle = AnVILFS(**props)
        return handle
__all__ = ('AnVILFilesSource',)


from io import BytesIO
from os import SEEK_END, SEEK_SET

from fs.enums import ResourceType
from fs.info import Info

from .baseresource import BaseAnVILResource


class BaseAnVILFile(BaseAnVILResource):
    def __init__(self, name, size, last_modified=None):
        self.name = name
        self.size = size
        self.last_modified = last_modified

    def getinfo(self):
        result = {
            "basic": {
                "name": self.name,
                "is_dir": False,
            },
            "details": {
                "type": ResourceType.file,
                "size": self.size,
                "modified": self.last_modified
            }
        }
        return Info(result)

    def string_to_buffer(self,string):
        buffer = BytesIO(string.encode('utf-8'))
        position = buffer.tell()
        buffer.seek(0, SEEK_END)
        self.size = buffer.tell()
        buffer.seek(position, SEEK_SET)
        return buffer

    def get_bytes_handler(self):
        raise NotImplementedError("Abstract method get_bytes_handler() not implemented")


from ..base import BaseAnVILResource, BaseAnVILFile, BaseAnVILFolder
from ..bucket import WorkspaceBucket, WorkspaceBucketFile
from ..namespace import Namespace
from ..reference import ReferenceDataFolder, ReferenceDataFile
from ..workspace import Workspace, WorkspaceData

from .testutils import BaseTest, FakeBlob

from fs.enums import ResourceType
from google.cloud import storage

# global test storage client
client = storage.Client()

# base anvil resource # -------------------------
class TestBaseANVILResource(BaseTest):
    def test_base_getinfo():
        DESC = "getinfo() is abstract"
        bar = BaseAnVILResource()
        bar.name = "test"
        try:
            bar.getinfo()
        except NotImplementedError as e:
            return BaseTest.success(DESC)
        return BaseTest.failure(DESC)

    def test_base_hash():
        DESC = "hash() hashes"
        bar = BaseAnVILResource()
        bar.name = "hasher"
        if bar.__hash__() == hash(("hasher", "BaseAnVILResource")):
            return BaseTest.success(DESC)
        else:
            return BaseTest.failure(DESC)


# base anvil file # -------------------------
class TestBaseAnVILFile(BaseTest):
    def test_bafile_getinfo():
        DESC = "getinfo()"
        name = "a"
        size = "1"
        last_mod = "lastmodstring"
        r_dict = {"basic": {
                    "name": name,
                    "is_dir": False,
                },
                "details": {
                    "type": ResourceType.file,
                    "size": size,
                    "modified": last_mod
                }
            }
        baf = BaseAnVILFile(name, size, last_mod)
        if baf.getinfo().raw == r_dict:
            return BaseTest.success(DESC)
        else:
            return BaseTest.failure(DESC)

    def test_bafile_get_bh():
        DESC = "open_bin() is abstract"
        baf = BaseAnVILFile("a", 1)
        try:
            baf.get_bytes_handler()
        except NotImplementedError as e:
            return BaseTest.success(DESC)
        return BaseTest.failure(DESC)

    def test_bafile_eq():
        DESC = "__eq__()"
        name = "a"
        size = 999
        last_mod = "yes"
        baf1 = BaseAnVILFile(name, size, last_mod)
        baf2 = BaseAnVILFile(name, size, last_mod)
        if baf1 == baf2:
            return BaseTest.success(DESC)
        else:
            return BaseTest.failure(DESC)

# base anvil folder # -------------------------
class TestBaseAnVILFolder(BaseTest):
    def test_bafolder_getinfo():
        DESC = "getinfo()"
        # folders end with '/' to differentiate between same-named files
        name = "a/"
        last_mod = "lastmodstring"
        r_dict = {
            "basic": {
                "name": name,
                "is_dir": True,
            },
            "details": {
                "type": ResourceType.directory,
                "modified": last_mod
            }
        }
        baf = BaseAnVILFolder(name, last_mod)
        if baf.getinfo().raw == r_dict:
            return BaseTest.success(DESC)
        else:
            return BaseTest.failure(DESC)

    def test_bafolder_hash():
        DESC = "__hash__()"
        name = "a/"
        last_mod = "yes"
        baf = BaseAnVILFolder(name, last_mod)
        if hash((name, "BaseAnVILFolder", last_mod)) == baf.__hash__():
            return BaseTest.success(DESC)
        else:
            return BaseTest.failure(DESC)

    def test_bafolder_get():
        DESC = "__getitem__()"
        baf = BaseAnVILFolder("A","B")
        bafile = BaseAnVILFile("X", 1)
        baf.children[bafile.name] = bafile
        if baf["X"] == bafile:
            return BaseTest.success(DESC)
        else:
            return BaseTest.failure(DESC)

    def test_bafolder_set():
        DESC = "__setitem__()"
        baf = BaseAnVILFolder("A","B")
        bafile = BaseAnVILFile("X", 1)
        baf[bafile.name] = bafile
        if bafile.name in baf.children:
            return BaseTest.success(DESC)
        else:
            return BaseTest.failure(DESC)

    def test_bafolder_gofp():
        DESC = "get_object_from_path()"
        baf = BaseAnVILFolder("Place","Holder")
        baf = BaseAnVILFolder("root")
        bafA = BaseAnVILFolder("A")
        bafB = BaseAnVILFolder("B")
        bafC = BaseAnVILFolder("C")
        bafD = BaseAnVILFile("D", 1)
        baf[bafA.name] = bafA
        bafA[bafB.name] = bafB
        bafB[bafC.name] = bafC
        bafC[bafD.name] = bafD
        fi = baf.get_object_from_path("A/B/C/D")
        fo = baf.get_object_from_path("A/B/C/")
        if fi == bafD and fo == bafC:
            return BaseTest.success(DESC)
        else:
            return BaseTest.failure(DESC)

# Namespace # -------------------------
class TestNamespace(BaseTest):
    def test_ns_init(ns_name, *args):
        DESC = "__init__()"
        ns = Namespace(ns_name)
        if ns.name != ns_name + "/":
            return BaseTest.failure(DESC + ": name")
        if ns.__hash__() != hash((ns_name+"/", "Namespace", None)):
            return BaseTest.failure(DESC + ": hash")
        return BaseTest.success(DESC)

    def test_ns_fetch_workspace(ns_name, ws_name):
        DESC = "fetch_workspace()"
        ns = Namespace(ns_name)
        ws = ns.fetch_workspace(ws_name)
        ws_folder_name = ws_name + "/"
        if ws.name[:-1] != ws_name:
            return BaseTest.failure(DESC + ": ws name")
        if ns[ws_folder_name] != ws:
            return BaseTest.failure(DESC + ": ws obj equivalence")
        return BaseTest.success(DESC)

# Workspace # --------------------------
class TestWorkspace(BaseTest):
    def test_ws_init(ns_name, ws_name):
        DESC = "__init__()"
        ns = Namespace(ns_name)
        ws = Workspace(ns, ws_name)
        if not isinstance(ws["Other Data/"]["Files/"], WorkspaceBucket):
            return BaseTest.failure(DESC + ": WorkspaceBucket not present")
        else:
            return BaseTest.success(DESC)

    def test_ws_fetchinfo(ns_name, ws_name):
        DESC = "fetch_api_info()"
        ns = Namespace(ns_name)
        ws = Workspace(ns, ws_name)
        info = ws.fetch_api_info(ws_name)
        if "workspace" not in info:
            return BaseTest.failure(DESC + ": fetched info incomplete")
        else:
            return BaseTest.success(DESC)

# WorkspaceBucketFile # ------------------
class TestWorkspaceBucketFile(BaseTest):
    def test_wsbucketfile_init():
        DESC = "__init__()"
        wsbf1 = WorkspaceBucketFile(FakeBlob())
        wsbf2 = WorkspaceBucketFile(FakeBlob("Steve", 666, "Never"))
        if not (
            wsbf1.name == FakeBlob.DEFAULT_NAME.split("/")[-1] and
            wsbf1.size == FakeBlob.DEFAULT_SIZE and
            wsbf1.last_modified == FakeBlob.DEFAULT_UPDATED
        ):
            return BaseTest.failure(DESC + ": Default object mismatch")
        elif not (
            wsbf2.name == "Steve" and
            wsbf2.size == 666 and
            wsbf2.last_modified == "Never"
        ):
            return BaseTest.failure(DESC + ": specified object mismatch")
        else:
            return BaseTest.success(DESC)

# WorkspaceBucket # ----------------------
class TestWorkspaceBucket(BaseTest):
    def test_bucket_init(bucket_name):
        DESC = "__init__()"
        wsb = WorkspaceBucket(bucket_name)
        if wsb.name != "Files/":
            return BaseTest.failure(DESC + ": name mismatch")
        else:
            return BaseTest.success(DESC)

    def test_bucket_insert(bucket_name):
        DESC = "insert_file()"
        wsb = WorkspaceBucket(bucket_name)
        wsb.insert_file(FakeBlob())
        wsbf = WorkspaceBucketFile(FakeBlob())
        fake_insert = wsb["afile/"]["in/"]["the/"]["bucket.nfo"]
        if fake_insert != wsbf:
            return BaseTest.failure(DESC + ": object mismatch")
        else:
            return BaseTest.success(DESC)

class TestWorkspaceData(BaseTest):
    def test__dict_to_buffer():
        DESC = "_dict_to_buffer()"
        data1 = {"a":"A", "b":"B"}
        data2 = {"b":"B", "a":"A"}
        bytesrep = b"a\tA\nb\tB\n"
        wsd = WorkspaceData("one", data1)
        buff1val = wsd.buffer.getvalue()
        wsd._dict_to_buffer(data2)
        buff2val = wsd.buffer.getvalue()
        if (buff1val == bytesrep and buff1val == buff2val):
            return BaseTest.success(DESC)
        else:
            return BaseTest.failure(DESC + ": identical values not equal after conversion")

    def test_data_init():
        DESC = "__init__()"
        data = {"space":"cowboy"}
        datab = b"space\tcowboy\n"
        wsd = WorkspaceData("maurice", data)
        if (wsd.name == "maurice" and 
            wsd.buffer.getvalue() == datab):
            return BaseTest.success(DESC)
        else:
            return BaseTest.failure(DESC + ": initialization failure")


class TestReferenceDataFile(BaseTest):
    def test_init():
        DESC = "__init__()"
        fake_blob = FakeBlob(name="a", size=2, updated="Never")
        rdf = ReferenceDataFile(fake_blob)
        if (rdf.blob_handle == fake_blob and
            rdf.name == "a" and
            rdf.size == 2 and
            rdf.last_modified == "Never" and
            not rdf.is_dir ):
            return BaseTest.success(DESC)
        else:
            return BaseTest.failure(DESC+ ": initialized object does not equal input")

    def test_make_rdfs():
        DESC = "make_rdfs()"
        # public reference gs bucket file:
        url = "gs://gcp-public-data--broad-references/hg38/v0/1000G_omni2.5.hg38.vcf.gz"
        bucket_name = "gcp-public-data--broad-references"
        bucket_path = "hg38/v0/1000G_omni2.5.hg38.vcf.gz"
        file_name = "1000G_omni2.5.hg38.vcf.gz"
        blob_size = 53238342
        blob_updated = "2019-12-06 23:55:16.264000+00:00"
        blob_name = bucket_path
        # get blob
        bucket = client.list_blobs(bucket_name, prefix=bucket_path)
        correct_blob = None
        for b in bucket:
            if b.name == bucket_path:
                correct_blob = {url: b}
                break
        r = ReferenceDataFile.make_rdfs(correct_blob)
        rdf = r[0]
        if (rdf == ReferenceDataFile(correct_blob[url]) and
            rdf.name == file_name and
            rdf.size == blob_size and
            str(rdf.last_modified) == str(blob_updated) and
            rdf.blob_handle == correct_blob[url] and
            not rdf.is_dir):
            return BaseTest.success(DESC)
        else:
            return BaseTest.failure(DESC + ": factory-constructed ReferenceDataFile != manually created")

class TestReferenceDataFolder(BaseTest):
    def test_init():
        url = "gs://gcp-public-data--broad-references/hg38/v0/1000G_omni2.5.hg38.vcf.gz"
        bucket_name = "gcp-public-data--broad-references"
        bucket_path = "hg38/v0/1000G_omni2.5.hg38.vcf.gz"
        file_name = "1000G_omni2.5.hg38.vcf.gz"
        file_size = 53238342
        # get blob
        bucket = client.list_blobs(bucket_name, prefix=bucket_path)
        correct_blob = None
        for b in bucket:
            if b.name == bucket_path:
                correct_blob = {url: b}
                break
        DESC = "__init__()"
        refs = {
            "source": {
                "reftype": correct_blob
                }}
        rdf = ReferenceDataFolder("test", refs)
        if (rdf.name == "test/" and
            isinstance(rdf["source/"]["reftype/"][file_name], ReferenceDataFile) and
            rdf["source/"]["reftype/"][file_name].name == file_name and
            rdf["source/"]["reftype/"][file_name].size == file_size):
            return BaseTest.success(DESC)
        else:
            return BaseTest.failure(DESC, "created ReferenceDataFile initialized in ReferenceDataFolder != remote source")


def run_all(anvil, files, folders):
    # define required arguments
    ns_name = anvil.namespace.name[:-1]
    ws_name = anvil.workspace.name[:-1]
    bucket_name = anvil.workspace.bucket_name
    print("=][=  [UNIT TESTING]  =][=")
    # format: a list of tuples
    #   [ (TestClassName, [list, of, arguments]), ... ]
    objs_n_args = [
        (TestBaseANVILResource, []),
        (TestBaseAnVILFile, []), 
        (TestBaseAnVILFolder, []),
        (TestNamespace, [ns_name, ws_name]),
        (TestWorkspace, [ns_name, ws_name]),
        (TestWorkspaceBucket, [bucket_name]),
        (TestWorkspaceBucketFile, []),
        (TestWorkspaceData, []),
        (TestReferenceDataFile, []),
        (TestReferenceDataFolder, [])
    ]
    results = [0, 0]
    failures = []
    for o in objs_n_args:
        _s, _f = o[0].run_tests(*o[1])
        if _f:
            failures.append(o[0].__name__ + " - {} failure(s)".format(_f))
        results[0] += _s
        results[1] += _f

    print(" GRAND TOTAL: {} FAILURES in {} TESTS\n".format(results[1], results[0] + results[1]))
    if failures:
        print("  failing modules:")
        for f in failures:
            print("\t"+f)
    return (results, failures)


from io import BytesIO
from os import SEEK_END, SEEK_SET

import re

import gs_chunked_io as gscio

from .basefile import BaseAnVILFile
from .basefolder import BaseAnVILFolder

class OtherDataFolder(BaseAnVILFolder):
    def __init__(self, attributes, bucket_name):
        super().__init__("Other Data")
        self.bucket_name = bucket_name
        self.attributes = attributes

    def lazy_init(self):
        workspacedata = dict(self.attributes)
        blocklist_prefixes = [
            "referenceData_",
            "description"
        ]
        for datum in self.attributes:
            for blocked in blocklist_prefixes:
                if datum.startswith(blocked):
                    del workspacedata[datum]
        if workspacedata:
            wsdf = WorkspaceDataFolder(workspacedata)
            #_wsd = WorkspaceData("WorkspaceData.tsv", workspacedata)
            self[wsdf.name] = wsdf
        _wsb = WorkspaceBucket(self.bucket_name)
        self[_wsb.name] = _wsb


class WorkspaceDataFolder(BaseAnVILFolder):
    def __init__(self, workspacedata):
        super().__init__("Workspace Data")
        self.workspacedata = workspacedata

    def lazy_init(self):
        files = {}
        for k in self.workspacedata:
            val = self.workspacedata[k]
            filetype = self.is_linkable_file(val)
            if filetype is not None:
                if filetype not in files:
                    files[filetype] = []
                files[filetype].append(val)
        linked_files = []
        for method in files:
            try:
                fresh_files = method.factory(files[method])
                linked_files.extend(fresh_files)
            except Exception as e:
                print(f"AnVILFS ERROR: SKIPPING FILE, could not be resolved due to the following error:")
                print(e)
                continue
        linked_files.append(WorkspaceData("WorkspaceData.tsv", self.workspacedata))
        for f in linked_files:
            self[f.name] = f


class WorkspaceBucketSubFolder(BaseAnVILFolder):
    def __init__(self, name, remaining, initializing_blob):
        super().__init__(name)
        self.remaining = remaining
        self.initializing_blob = initializing_blob

    def lazy_init(self):
        if len(self.remaining) == 1:
            self[self.remaining[0]] = WorkspaceBucketFile(self.initializing_blob)
            del self.initializing_blob
        else:
            subname = self.remaining[0] + '/'
            self[subname] = WorkspaceBucketSubFolder(subname, self.remaining[1:], self.initializing_blob)

class WorkspaceBucket(BaseAnVILFolder):
    def __init__(self, bucket_name):
        super().__init__("Files")
        self.bucket_name = bucket_name

    def lazy_init(self):
        google_bucket = self.gc_storage_client.get_bucket(self.bucket_name)
        blobs = google_bucket.list_blobs()
        self.initialized = True
        for blob in blobs:
            self.insert_file(blob)

    def insert_file(self, bucket_blob):
        # name relative to the path from workspace bucket
        path = bucket_blob.name
        if path[-1] == "/":
            raise Exception("Files should be set, not folders")
        s = path.split("/")
        if len(s) == 1:
            _wsbf = WorkspaceBucketFile(bucket_blob)
            self[_wsbf.name] = _wsbf
        else:
            subname = s[0]+'/'
            self[subname] = WorkspaceBucketSubFolder(subname, s[1:], bucket_blob)


class WorkspaceBucketFile(BaseAnVILFile):
    def __init__(self, blob):
        self.name = blob.name.split("/")[-1]
        self.size = blob.size
        self.last_modified = blob.updated
        self.blob_handle = blob
        self.is_dir = False
    
    def get_bytes_handler(self):
        return gscio.Reader(self.blob_handle)


class WorkspaceData(BaseAnVILFile):
    def __init__(self, name, data_dict):
        self.name = name
        self.buffer = self._dict_to_buffer(data_dict)
        self.last_modified = None
    
    def _dict_to_buffer(self, d):
        # only keys that match the below regex are valid 
        keys = [k for k in d.keys() if bool(re.match("^[A-Za-z0-9_-]*$", k)) ]
        data = ""
        for k in keys:
            data += f"{k}\t{d[k]}\n"
        buffer = BytesIO(data.encode('utf-8'))
        position = buffer.tell()
        buffer.seek(0, SEEK_END)
        self.size = buffer.tell()
        buffer.seek(position, SEEK_SET)
        return buffer

    def get_bytes_handler(self):
        return self.buffer


from inspect import getfullargspec

class BaseTest:
    TEST_INDENT_CHARS=24 #(3 tabs)
    def success(msg, details=None):
        if details:
            msg = msg + f": {details}"
        print(" - SUCCESS: {}".format(msg))
        return (1, 0)

    def failure(msg, details=None):
        if details:
            msg = msg + f": {details}"
        print(" - FAILURE: {}".format(msg))
        return (0, 1)

    @classmethod
    def run_tests(self, *args):
        name = self.__name__[4:] if self.__name__[0:4] == "Test" else self.__name__ 
        tests = self.get_tests()
        print("[ {} ]:\n tests: {}".format(name, tests))
        s = 0
        f = 0
        for t in tests:
            gap = 24 - len(t) + 3
            print(" - {}:{}\t".format(t, " "*gap), end='')
            arg_ct = len(getfullargspec(getattr(self, t)).args)
            # here because of "'NoneType' object is not iterable"?
            #   did you RETURN BaseTest.success/failure?
            _s, _f = getattr(self, t)(*args[0:arg_ct])
            s += _s
            f += _f
        print("\tTOTAL: {} FAILURES out of {} TESTS\n".format(f, s+f))
        return (s,f)

    @classmethod
    def get_tests(self):
        return [f for f in dir(self) if callable(getattr(self, f)) and f[0:5] == "test_" ]

class FakeBlob:
    DEFAULT_NAME = "afile/in/the/bucket.nfo"
    DEFAULT_SIZE = 1
    DEFAULT_UPDATED = "some time"

    def __init__(self, name=DEFAULT_NAME, size=DEFAULT_SIZE, updated=DEFAULT_UPDATED):
        self.name = name
        self.size = size
        self.updated = updated

    def __eq__(self, second):
        return (self.name == second.name
                and self.size == second.size
                and self.updated == second.updated)



# python-compatible syntax to define required vars:
# 
# namespace_name: <your AnVIL Google project name>
#    NOTE: must have google cloud sdk configured to this project
# workspace_name: <the target AnVIL workspace name>
config = {
    "namespace_name": "<some project>",
    "workspace_name": "<some workspace>"
}


from .clientrepository import ClientRepository
from .basefolder import BaseAnVILFolder
from .google import DRSAnVILFile
from .namespace import Namespace
from .workspace import Workspace
from .workloadidentitycredentials import WorkloadIdentityCredentials

from google.auth.transport.requests import AuthorizedSession

from fs.base import FS
from fs.errors import DirectoryExpected, ResourceNotFound, FileExpected


class AnVILFS(FS, ClientRepository):
    DEFAULT_API_URL = "https://api.firecloud.org/api/"
    #DEV_API_URL="https://firecloud-orchestration.dsde-dev.broadinstitute.org/api/"

    def __init__(self, namespace, workspace, api_url=None, on_anvil=False, drs_url=None):
        super(AnVILFS, self).__init__()
        ClientRepository.base_project = namespace
        if not api_url:
            api_url = self.DEFAULT_API_URL
        if drs_url:
            DRSAnVILFile.api_url = drs_url
        if on_anvil:
            scopes = ['https://www.googleapis.com/auth/userinfo.email', 'https://www.googleapis.com/auth/userinfo.profile', 'https://www.googleapis.com/auth/cloud-platform']
            credentials = WorkloadIdentityCredentials(scopes=scopes)
            self.fapi.__setattr__("__SESSION", AuthorizedSession(credentials))
            self.fapi.fcconfig.set_root_url(api_url)
        self.namespace = Namespace(namespace)
        self.workspace = self.namespace.fetch_workspace(workspace)
        self.rootobj = self.workspace  # leaving the option to make namespace root

    def getinfo(self, path, namespaces=None):
        return self.rootobj.get_object_from_path(path).getinfo()

    # Get a list of resource names (str) in a directory.
    def listdir(self, path):
        if path == "/" or path == "":
            return self.rootobj.keys()
        try:
            maybe_dir = self.rootobj.get_object_from_path(path)
        except KeyError as ke:
            raise ResourceNotFound("Resource {} not found".format(path))
        if isinstance(maybe_dir, BaseAnVILFolder):
            return maybe_dir.keys()
        else:
            raise DirectoryExpected(f"{path}")

    def scandir(self, path):
        if path[-1] != "/":
            path = path + "/"
        result = []
        l = self.listdir(path)
        for o in l:
            result.append(self.getinfo(path+o))
        return result

    def makedir():  # Make a directory.
        raise Exception("makedir not implemented")

    def openbin(self, path, mode="r", buffering=-1, **options):
        obj = self.rootobj.get_object_from_path(path)
        try:
            return obj.get_bytes_handler()
        except AttributeError as e:
            raise FileExpected(
                "Error: requested object is not a file:\n  {}".format(path))

    def remove():  # Remove a file.
        raise Exception("remove not implemented")

    def removedir():  # Remove a directory.
        raise Exception("removedir not implemented")

    def setinfo():  # Set resource information.
        raise Exception("setinfo not implemented")
    # for network systems, scandir needed otherwise default calls a combination of listdir and getinfo for each file.
